/******************************************************************************
*
*  FILENAME :	perms.c
*
*  DESCRIPTION :
*
*	Source code used to determine EFAC permissions.
*
*  PUBLIC FUNCTIONS :
*
*	void get_names(char (*names)[MAX_NAME_LEN+1])
*	uint get_perms(Boolean print_info)
*	void print_accset(uint perm_num, char *tag, Boolean perm_given)
*
*  NOTES :
*
*	Written by Paul Chesson (1994)
*	E-Mail: chesson@mundil.cs.mu.OZ.AU
*
*/


/******************************************************************************
*  INCLUDE FILES
******************************************************************************/

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <search.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>

#include "bugbyte.h"
#include "efacs.h"

#include "file.h"
#include "logic.h"
#include "pattern.h"
#include "perms.h"


/******************************************************************************
*  PUBLIC DECLARATIONS
******************************************************************************/

/* evaluation functions for log_eval() */
Boolean name_eval(char *name, void *data);
Boolean file_eval(char *pattern, void *data);

void print_accset(uint perm_num, char *tag, Boolean perm_given);


/******************************************************************************
*  PRIVATE DECLARATIONS
******************************************************************************/

/* ----- context ----------------------------------------------------------- */

/* file access classes */

#define USER	(0)
#define GROUP	(1)
#define OTHER	(2)

/* ----- function prototypes ----------------------------------------------- */

static uint acl_perms(const char *filename, const char *acl_file, 
		      Boolean print_info);
static uint auto_perms(const char *filename, Boolean print_info);
static uint continue_readline(char *buf, const uint bufsiz, FILE *stream,
			      char *err_id);

/* ----- macros ------------------------------------------------------------ */

/*
 *	ADD_NAME - Assign NAME to names[] and hash table.
 *		   Increment num_names.
 */

#define ADD_NAME(NAME, UNIX_GROUP)					\
{									\
    Boolean true = TRUE;						\
    assert(num_names+1 < MAX_NAMES);					\
    strncpy(names[num_names], (NAME), MAX_NAME_LEN);			\
    item.key = names[num_names++];					\
    hsearch(item, ENTER);						\
    if (UNIX_GROUP == true)		/* keeps lint happy	*/	\
    {									\
	strcpy(names[num_names], "_");					\
	strcat(names[num_names], (NAME));				\
	item.key = names[num_names++];					\
	hsearch(item, ENTER);						\
    };									\
}


/******************************************************************************
*  PUBLIC FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	void get_names(char (*names)[MAX_NAME_LEN+1])
*
*  DESCRIPTION :
*
*	Creates a set of names and aliases which the user has access to
*	when determining EFAC permissions.
*
*  INPUTS :
*
*	PARAMETERS:	names[]		empty array of pointers to names
*		
*  OUTPUTS :
*
*	PARAMETERS:	names[]		array of names for user
*		
*	GLOBALS:	???		hash table for names[]
*		
*/

void get_names(char (*names)[MAX_NAME_LEN+1])
{
    const uid_t	uid = getuid();		/* efac user's user id		*/
    const uid_t	euid = geteuid();	/* efacs owner's user id	*/
    gid_t	pw_gid;			/* efac user's default group id	*/
    char	alias_file[FILENAME_MAX+1];  /* path name of alias file	*/
    uint	line = 0;		/* current line in alias file	*/
    uint	num_names = 0;		/* count of names for user	*/
    struct group *grp;
    ENTRY	item;
    FILE	*fp;
    char	buf[BUFSIZ];

    /* _____ add user name to list _____ */

    hcreate(MAX_NAMES);
    ADD_NAME(getpwuid(uid)->pw_name, FALSE);

    /* _____ add unix group names to list _____ */

    pw_gid = getpwuid(uid)->pw_gid;
    while ((grp = getgrent()) != NULL)
    {
        char	**mem;

	/* _____ check for default gid _____ */

	if (grp->gr_gid == pw_gid)
	{
	    ADD_NAME(grp->gr_name, TRUE);
	    continue;
	}

	/* _____ check for group membership _____ */

	mem = grp->gr_mem;
	while (*mem != NULL)
	{
	    if (STR_EQ((const char *)*mem, (const char *)names[0]))
	    {
	        ADD_NAME(grp->gr_name, TRUE);
	        break;
	    }
	    mem++;
	}

    }  /* while more entries in /etc/group */

    /*
     *	    Add EFACS aliases to list.
     */

    /* _____ open aliases file _____ */

    strcpy(alias_file, getpwuid(euid)->pw_dir);
    strcat(alias_file, ALIAS_PATH);

    if ((fp = fopen(alias_file, "r")) == NULL)
	return;

    /* _____ read line from aliases file _____ */

    while (memset(buf, NUL, BUFSIZ) && fgets(buf, BUFSIZ, fp) != NULL)
    {
	char	*exp;		/* expression defining alias membership	*/
	char	err_id[BUFSIZ]; /* error identification string		*/

	/* _____ show errors if running own server _____ */

	if (uid == euid)
	    sprintf(err_id, "%s:%u", alias_file, ++line);
	else
	    err_id[0] = NUL;

	/* _____ handle continuations, blank, and comment lines _____ */

	line += continue_readline(buf, BUFSIZ, fp, err_id);
	if (*buf == '\n' || *buf == '#')
	    continue;

	/* _____ evaluate alias expression _____ */

	strtok(buf, " \t");
	exp = buf + strlen(buf) + 1;
	switch ( log_eval(strtok(exp, "\n"), err_id,
		(Boolean (*)(char *, void *))name_eval, NULL) )
	{
	    case TRUE:
		ADD_NAME(buf, FALSE);
		break;
	    case FALSE:
		break;
	    case BLANK:
		fprintf(stderr, "%s: missing logical expression\n", err_id);
		break;
	    default:
		assert(FALSE);
	}

    }  /* while there is another line in the alias file */

    close_file(alias_file, fp);

}


/******************************************************************************
*
*  NAME :	uint get_perms(const char *filename, Boolean print_info)
*
*  DESCRIPTION :
*
*	Compute the EFAC permissions for a file.
*
*  INPUTS :
*
*	PARAMETERS:	filename	name of file being checked
*			print_info	whether to display its access sets
*		
*  OUTPUTS :
*
*	RETURN:		Bit pattern describing available permissions.
*
*  NOTES :
*
*	See perms.h for names of flags in output pattern.
*
*/

uint get_perms(const char *filename, Boolean print_info)
{
    char	dir_acl[FILENAME_MAX+1];	/* directory acl file	*/
    char	file_acl[FILENAME_MAX+1];	/* file's own acl file	*/
    uint	perms;				/* returned permissions	*/

    strcpy(dir_acl, get_dir(filename));
    strcat(dir_acl, ".acl");
    strcpy(file_acl, filename);
    if (strlen(filename) < 4
	|| STR_DIFF(filename + strlen(filename) - 4, ".acl"))
    {
	strcat(file_acl, ".acl");
    }

    perms = ( acl_perms(filename, dir_acl, print_info)
	     | acl_perms(filename, file_acl, print_info)
	     | auto_perms(filename, print_info) );
    if ((perms & RW_ACCESS) == RW_ACCESS)
	perms |= A_ACCESS;
    
    return perms;
}


/******************************************************************************
*
*  NAME :	void print_accset(uint perm_num, char *tag, Boolean perm_given)
*
*  DESCRIPTION :
*
*	Display an access set.  EFAC permissions and user expression.
*
*  INPUTS :
*
*	PARAMETERS:	perm_num	access bit string
*			tag		user expression field string
*			perm_given	if user belongs to access set
*		
*  NOTES :
*
*	Permission flag characters shown in upper-case if perm_given.
*	See perms.h for names of flags in output pattern.
*
*/

void print_accset(uint perm_num, char *tag, Boolean perm_given)
{
    int		case_change = ('A'-'a') * perm_given;

    while (*tag == SPC || *tag == TAB)
	tag++;

    if (perm_num)
    {
	printf("%c%c%c%c%c\t%s\n",
	       (perm_num & A_ACCESS) ? 'a'+case_change : '-',
	       (perm_num & C_ACCESS) ? 'c'+case_change : '-',
	       (perm_num & I_ACCESS) ? 'i'+case_change : '-',
	       (perm_num & R_ACCESS) ? 'r'+case_change : '-',
	       (perm_num & W_ACCESS) ? 'w'+case_change : '-',
	       tag);
    }
}


/******************************************************************************
*
*  NAME :	Boolean name_eval(char *name, void *data)
*
*  DESCRIPTION :
*
*	Indicates if an EFAC name token is in the names[] array.
*
*  INPUTS :
*
*	PARAMETERS:	name		EFAC name token
*			data		(not used)
*
*	GLOBAL:		???		hash table for names[]
*		
*  OUTPUTS :
*
*	RETURN:		TRUE		token present in array
*			FALSE		otherwise
*
*/

Boolean name_eval(char *name, void *data)
{
    ENTRY	item;

    data = data;		/* keeps lint happy for unused input	*/
    item.key = name;
    return (hsearch(item, FIND) != NULL) ? TRUE : FALSE;
}


/******************************************************************************
*
*  NAME :	Boolean file_eval(char *pattern, void *data)
*
*  DESCRIPTION :
*
*	Indicates if file pattern token matches name of file requested.
*
*  INPUTS :
*
*	PARAMETERS:	name		EFAC name token
*			data		requested filename
*
*  OUTPUTS :
*
*	RETURN:		TRUE		filename matches pattern
*			FALSE		otherwise
*
*  NOTES :
*
*	Ignore any preceding path of filename.
*
*/

Boolean file_eval(char *pattern, void *data)
{
    char *cp = (char *)data + strlen((char *)data);

    while (cp != (char *)data && *(cp-1) != '/')
	cp--;
    return pat_match(cp, pattern);
}


/******************************************************************************
*  PRIVATE FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	uint acl_perms(const char *filename, const char *acl_file, 
*			       Boolean print_info)
*
*  DESCRIPTION :
*
*	Compute the EFAC permissions for a file from a single acl_file only.
*
*  INPUTS :
*
*	PARAMETERS:	filename	name of file being checked
*			acl_file	name of acl file used to check
*			print_info	whether to display access sets
*		
*  OUTPUTS :
*
*	RETURN:		Bit pattern describing available permissions.
*
*  NOTES :
*
*	See perms.h for names of flags in output pattern.
*
*/

static uint acl_perms(const char *filename, const char *acl_file,
		      Boolean print_info)
{
    uint	perms = 0;		/* total EFAC permissions for acl  */
    int		line = 0;		/* current line number of acl file */
    Boolean	inside_block = FALSE;	/* if looking inside access block  */
    Boolean	file_match = FALSE;	/* if access block header...       */
					/*   matches requested file        */
    const uid_t	uid = getuid();		/* user executing efacs            */
    const uid_t	euid = geteuid();	/* user owning efacs               */
    struct stat	statbuf;
    FILE	*fp;
    char	buf[BUFSIZ];

    /*
     *	    Return if access list does not exist or is not owned by
     *	    the owner of efacs being executed.
     */

    if ( (fp = fopen(acl_file, "r")) == NULL 
	|| stat(acl_file, &statbuf) < 0
	|| statbuf.st_uid != euid )
    {
	return 0;
    }

    /* _____ search acl file for matching permission _____ */

    while (memset(buf, '\0', BUFSIZ) && fgets(buf, BUFSIZ, fp) != NULL)
    {
	char	err_id[BUFSIZ];

	/* _____ print errors only if user owns server _____ */

	if (print_info == FALSE && uid == euid)
	    sprintf(err_id, "%s:%d", acl_file, ++line);
	else
	    err_id[0] = '\0';

	line += continue_readline(buf, BUFSIZ, fp, err_id);

	/* _____ new lines terminate blocks _____ */

	if (*buf == '\n')
	{
	    inside_block = FALSE;
	    continue;
	}

	/* _____ ignore comments _____ */

	if (*buf == '#')
	    continue;

	/* _____ at start of block, check if file matches expression _____ */ 

	if (!inside_block)
	{
	    switch(log_eval(strtok(buf, "\n"), err_id,
	        (Boolean (*)(char *, void *))file_eval, (void *)filename))
	    {
		case BLANK:
		    continue;
		case TRUE:
		    file_match = TRUE;
		    inside_block = TRUE;
		    break;
		case FALSE:
		case ERROR:
		    file_match = FALSE;
		    inside_block = TRUE;
		    break;
	    }
	}

	/* _____ ignore lines inside block not applicable to the file _____ */

	else if (inside_block && !file_match)
	    continue;
	
	/* _____ when inside a block pertaining to the requested file _____ */

	else if (inside_block && file_match)
	{
	    char	*exp;		/* access set user expression	 */
	    uint	line_perms;	/* permissions for current line	 */
	    Boolean	result;		/* if user belongs to access set */

	    strtok(buf, " \t");
	    exp = buf + strlen(buf) + 1;

	    result = log_eval(strtok(exp, "\n"), err_id,
			      (Boolean (*)(char *, void *))name_eval, NULL);

	    line_perms =
		(  ((strchr(buf, 'a') != NULL) ? A_ACCESS : 0)
		 | ((strchr(buf, 'c') != NULL) ? C_ACCESS : 0)
		 | ((strchr(buf, 'i') != NULL) ? I_ACCESS : 0)
		 | ((strchr(buf, 'r') != NULL) ? R_ACCESS : 0)
		 | ((strchr(buf, 'w') != NULL) ? W_ACCESS : 0)
		 | ((strchr(buf, 'w') != NULL)
		    && (strchr(buf, 'r') != NULL) ? A_ACCESS : 0) );

	    if (result == TRUE || result == BLANK)
		perms |= line_perms;

	    if (print_info)
	    {
		print_accset(line_perms, exp,
			     (result == TRUE || result == BLANK));
	    }
	}

	else
	    assert(FALSE);

    }  /* while there is another block in the access file */

    close_file(acl_file, fp);

    return perms;

}


/******************************************************************************
*
*  NAME :	uint auto_perms(const char *filename, Boolean print_info)
*
*  DESCRIPTION :
*
*	Compute the EFAC automatic permissions for a file.
*	These are derived from the operations achievable under the
*	current normal permissions for the file.
*
*  INPUTS :
*
*	PARAMETERS:	filename	name of file being checked
*			print_info	whether to display access sets
*		
*  OUTPUTS :
*
*	RETURN:		Bit pattern describing available permissions.
*
*  NOTES :
*
*	See perms.h for names of flags in output pattern.
*
*/

static uint auto_perms(const char *filename, Boolean print_info)
{
    struct stat	f_statbuf;		/* requested file stats		*/
    struct stat	d_statbuf;		/* file's directory's stats	*/
    const uid_t	uid = getuid();		/* efacs caller's user id	*/
    const uid_t	euid = geteuid();	/* efacs owner's user id	*/
    const gid_t	gid = getgid();		/* efacs caller's group id	*/
    uint	u_perms = 0;		/* user EFAC permissions	*/
    uint	g_perms = 0;		/* group EFAC permissions	*/
    uint	o_perms = 0;		/* other EFAC permissions	*/
    uint	n_perms = 0;		/* EPACS permissions on...	*/
					/*   non-existant files		*/
    uint	perms = 0;		/* total EFAC permissions	*/
    uint	dir_mask = 0;		/* directory bit mask		*/
    char	tag[MAX_LINE_LEN];	/* user expression string	*/
    short	file_access = USER;	/* normal file access level	*/

    if (stat(get_dir(filename), &d_statbuf) < 0)
    {
	perror(get_dir(filename));
	return perms;
    }

    /*
     *	    Auto permissions on possibly non-existant files, when running
     *	    your own server.  Creation permission if you have write and
     *	    access permissions to the directory.  Info permissions always.
     */

    if (uid == euid)
    {
	n_perms |= I_ACCESS;
	if (d_statbuf.st_uid == getuid())
	    dir_mask = S_IWUSR | S_IXUSR;
	else if (d_statbuf.st_gid == gid)
	    dir_mask = S_IWGRP | S_IXGRP;
	else
	    dir_mask = S_IWOTH | S_IXOTH;

	if ((d_statbuf.st_mode & dir_mask) == dir_mask)
	{
	    n_perms |= C_ACCESS;
	    if (print_info)
	        print_accset(n_perms, getpwuid(uid)->pw_name, TRUE);
	}
	perms |= n_perms;
    }

    /*
     *	    Auto permissions on existing files, based on their normal
     *	    permissions.
     */

    if (stat(filename, &f_statbuf) < 0)
	return perms;

    /* _____ determine user, group and other permissions _____ */

    if (f_statbuf.st_mode & S_IRUSR)
	u_perms |= R_ACCESS;
    if (f_statbuf.st_mode & S_IWUSR)
	u_perms |= W_ACCESS;
    if ((u_perms & RW_ACCESS) == RW_ACCESS)
	u_perms |= A_ACCESS;

    if (f_statbuf.st_mode & S_IRGRP)
	g_perms |= R_ACCESS;
    if (f_statbuf.st_mode & S_IWGRP)
	g_perms |= W_ACCESS;
    if ((g_perms & RW_ACCESS) == RW_ACCESS)
	g_perms |= A_ACCESS;

    if (f_statbuf.st_mode & S_IROTH)
	o_perms |= R_ACCESS;
    if (f_statbuf.st_mode & S_IWOTH)
	o_perms |= W_ACCESS;
    if ((o_perms & RW_ACCESS) == RW_ACCESS)
	o_perms |= A_ACCESS;

    /* _____ apply appropriate permissions _____ */

    if (f_statbuf.st_uid == euid)
    {
	if (f_statbuf.st_uid == uid)
	{
	    file_access = USER;
	    perms |= u_perms;
	    dir_mask = S_IRUSR | S_IXUSR;
	}
	else if (f_statbuf.st_gid == gid)
	{
	    file_access = GROUP;
	    perms |= g_perms;
	    dir_mask = S_IRGRP | S_IXGRP;
	}
	else
	{
	    file_access = OTHER;
	    perms |= o_perms;
	    dir_mask = S_IROTH | S_IXOTH;
	}
    }

    /* _____ print access sets if directory readable and accessable _____ */

    if (print_info && (d_statbuf.st_mode & dir_mask) == dir_mask
	&& f_statbuf.st_uid == euid)
    {
	print_accset(u_perms, getpwuid(f_statbuf.st_uid)->pw_name,
		     (file_access == USER));
	sprintf(tag, "!%s && _%s",
		getpwuid(f_statbuf.st_uid)->pw_name,
		getgrgid(f_statbuf.st_gid)->gr_name);
	print_accset(g_perms, tag, (file_access == GROUP));
	sprintf(tag, "!%s && !_%s",
		getpwuid(f_statbuf.st_uid)->pw_name,
		getgrgid(f_statbuf.st_gid)->gr_name);
	print_accset(o_perms, tag, (file_access == OTHER));
    }

    return perms;
}


/******************************************************************************
*
*  NAME :	uint continue_readline(char *buf, const uint bufsiz,
*				       FILE *stream, char *err_id)
*
*  DESCRIPTION :
*
*	Tests if a line is terminated by an escaped (\) newline, and if so,
*	continues reading further lines, appending them to the first.
*
*  INPUTS :
*
*	PARAMETERS:	buf		buffer containing line of text read
*			bufsiz		maximum size of buffer
*			stream		file stream to continue reading on
*			err_id		error identification tag to use
*		
*  OUTPUTS :
*
*	PARAMETERS:	buf		buffer containg full line of text
*
*	RETURN:		Number of extra lines read in.
*
*/

static uint continue_readline(char *buf, const uint bufsiz, FILE *stream,
			      char *err_id)
{
    uint	cont_lines = 0;	/* number of lines of continuation	*/

    while (strlen(buf) >= 2 && STR_EQ(buf + strlen(buf) - 2, "\\\n"))
    {
	int chars_left = bufsiz - strlen(buf);
	char *buf2;
	
	MALLOC_TYPE(buf2, char, chars_left);
	if (fgets(buf2, chars_left, stream) == NULL)
	{
	    fprintf(stderr, "%s: no line after '\\'\n", err_id);
	    exit(-1);
	}
	cont_lines++;
	buf[strlen(buf)-2] = SPC;
	buf[strlen(buf)-1] = NUL;
	strcat(buf, buf2);
	free(buf2);
    }
    return cont_lines;
}

